import unittest

from pyramid.exceptions import ConfigurationError

from kinto_fxa.utils import parse_clients


class UtilsTest(unittest.TestCase):
    def test_parse_resources_raises_configuration_error_in_case_of_conflict(self):

        settings = {}
        settings['fxa-oauth.oauth_uri'] = 'https://oauth.accounts.firefox.com/v1'
        settings['fxa-oauth.cache_ttl_seconds'] = '0.01'
        settings['fxa-oauth.clients.notes.client_id'] = 'c73e46074a948932'
        settings['fxa-oauth.clients.notes.required_scope'] = (
            'https://identity.mozilla.org/apps/notes')
        settings['fxa-oauth.clients.lockbox.client_id'] = '299062f8b3838932'
        settings['fxa-oauth.clients.lockbox.required_scope'] = (
            'https://identity.mozilla.org/apps/notes')

        self.assertRaises(ConfigurationError, parse_clients, settings)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Script to process account-related events from an SQS queue.

This script polls an SQS queue for events indicating activity on an upstream
account, as documented here:

  https://github.com/mozilla/fxa-auth-server/blob/master/docs/service_notifications.md

The following event types are currently supported:

  * "delete": the account was deleted; we delete their default bucket
    to comply with GDPR. Note that this won't be sufficient in
    applications where users can store data in locations besides the
    default bucket.

Note that this script may not be necessary in all applications of
kinto-fxa.

"""

import itertools
import json
import logging
import re
import uuid

import boto3
from ec2_metadata import ec2_metadata
from kinto.core.utils import hmac_digest
import transaction as current_transaction

logger = logging.getLogger(__name__)


def get_default_bucket_id(config, uid):
    secret = config['registry'].settings['userid_hmac_secret']
    digest = hmac_digest(secret, uid)
    return str(uuid.UUID(digest[:32]))


def process_account_events(config, queue_name, aws_region=None,
                           queue_wait_time=20):
    """Process account events from an SQS queue.

    This function polls the specified SQS queue for account-related events,
    processing each as it is found.  It polls indefinitely and does not return;
    to interrupt execution you'll need to e.g. SIGINT the process.
    """
    logger.info("Processing account events from %s", queue_name)
    statsd = getattr(config['registry'], 'statsd', None)
    process_one = process_account_event
    if statsd:
        process_one = statsd.timer("process_account_event")(process_one)
    try:
        # Connect to the SQS queue.
        # If no region is given, infer it from the instance metadata.
        if aws_region is None:
            logger.debug("Finding default region from instance metadata")
            aws_region = ec2_metadata.region

        logger.debug("Connecting to queue %r in %r", queue_name, aws_region)
        sqs = boto3.resource('sqs', region_name=aws_region)
        queue = sqs.get_queue_by_name(QueueName=queue_name)

        # Poll for messages indefinitely.
        # Use a wacky looping construct that can be mocked in tests.
        for x in itertools.count():
            msgs = queue.receive_messages(WaitTimeSeconds=queue_wait_time)
            for msg in msgs:
                process_one(config, msg.body)
                # This intentionally deletes the event even if it was some
                # unrecognized type.  No point leaving a backlog.
                msg.delete()

    except Exception:
        logger.exception("Error while processing account events")
        raise


def process_account_event(config, body):
    """Parse and process a single account event."""
    registry = config['registry']
    settings = registry.settings
    storage = registry.storage
    permission = registry.permission

    # Try very hard not to error out if there's junk in the queue.
    try:
        # Messages are a string of JSON, which, when parsed, has a
        # Message field, which is a string of JSON that actually
        # contains what we want.
        event = json.loads(body)
        event = json.loads(event['Message'])
        event_type = event["event"]
        uid = event["uid"]
    except (ValueError, KeyError) as e:
        logger.exception("Invalid account message: %r", e)
    else:
        if event_type == "delete":
            # Delete everything from storage and permissions for
            # this user.
            logger.info("Processing account delete for %r", uid)

            # Go through configured policies to find the policy name.
            prefix = ""
            for k, v in settings.items():
                m = re.match('multiauth\\.policy\\.(.*)\\.use', k)
                if m:
                    if v.endswith('FxAOAuthAuthenticationPolicy'):
                        prefix = "{}:".format(m.group(1))

            userids = [prefix + uid]
            # Go through configured clients.
            for k, v in settings.items():
                m = re.match('fxa-oauth\\.clients\\.(.*)\\.client_id', k)
                if m:
                    suffix = "-{}".format(m.group(1))
                    userids += [prefix + uid + suffix]

            for uid in userids:
                default_bucket_id = get_default_bucket_id(config, uid)
                bucket_uri = '/buckets/{}'.format(default_bucket_id)
                logger.info('Deleting bucket %r', bucket_uri)
                # Delete the bucket and all its descendants.
                # This code is similar to that from kinto.views.buckets:on_buckets_deleted.
                for parent_id in [bucket_uri, bucket_uri + '/*']:
                    storage.delete_all(
                        parent_id=parent_id,
                        collection_id=None,
                        with_deleted=False,
                    )
                    # Purge tombstones too.
                    storage.purge_deleted(
                        parent_id=parent_id,
                        collection_id=None,
                    )
                    permission.delete_object_permissions(parent_id)
            current_transaction.commit()
        else:
            logger.warning("Dropping unknown event type %r",
                           event_type)



from pyramid.exceptions import ConfigurationError
from collections import OrderedDict


def fxa_conf(request, name):
    key = 'fxa-oauth.%s' % name
    return request.registry.settings[key]


def parse_clients(settings):
    resources = OrderedDict()
    scope_routing = {}

    for setting_key, setting_value in settings.items():
        if not setting_key.startswith('fxa-oauth.'):
            continue
        elif setting_key.startswith('fxa-oauth.clients.'):
            parts = setting_key.split('.', 4)
            client_name, setting_basename = parts[2:]
        else:
            _, setting_basename = setting_key.split('.', 1)
            client_name = "default"

        if setting_basename == 'required_scope':
            if setting_value in scope_routing:
                message = '{} is already required for another config.'.format(setting_value)
                raise ConfigurationError(message)
            if setting_value is not None:
                scope_routing[setting_value] = client_name

        resource = resources.setdefault(client_name, OrderedDict())
        resource[setting_basename] = setting_value

    return resources, scope_routing



import mock
import unittest

from kinto_fxa.scripts import __main__ as main


class TestScripts(unittest.TestCase):
    def setUp(self):
        process_account_events_patcher = mock.patch(
            'kinto_fxa.scripts.__main__.process_account_events')
        self.process_account_events = process_account_events_patcher.start()
        self.addCleanup(process_account_events_patcher.stop)

        bootstrap_patcher = mock.patch('kinto_fxa.scripts.__main__.bootstrap')
        self.bootstrap = bootstrap_patcher.start()
        self.addCleanup(bootstrap_patcher.stop)

        logging_patcher = mock.patch('logging.config.fileConfig')
        self.fileConfig = logging_patcher.start()
        self.addCleanup(logging_patcher.stop)

        self.config = mock.Mock()
        self.bootstrap.return_value = self.config

    def test_call_main(self):
        main.main(["process-account-events", "my-queue-name"])
        self.bootstrap.assert_called_with(main.DEFAULT_CONFIG_FILE)
        self.process_account_events.assert_called_with(
            self.config, 'my-queue-name', None, 20
        )
        self.fileConfig.assert_called_with(main.DEFAULT_CONFIG_FILE,
                                           disable_existing_loggers=False)



import json
import mock
import unittest

from kinto_fxa.scripts.process_account_events import (
    get_default_bucket_id,
    process_account_event,
    process_account_events,
)


class TestProcessAccountEvent(unittest.TestCase):
    def setUp(self):
        self.registry = mock.Mock()
        self.registry.settings = {
            'userid_hmac_secret': 'efghi'
        }
        self.config = {"registry": self.registry}
        self.uid = 'abcd'
        # Computed this by hand in the Python terminal
        self.bucket_id = '719b8628-0230-7b89-46e8-a365d68a66d1'
        self.real_message = json.dumps({
            "Message": json.dumps({
                "event": "delete",
                "uid": self.uid,
            })
        })

    @mock.patch('kinto_fxa.scripts.process_account_events.logger')
    def test_message_that_isnt_json_is_dropped(self, logger):
        process_account_event(self.config, "blah")
        calls = logger.exception.call_args_list
        self.assertEqual(len(calls), 1)
        args = calls[0][0]
        self.assertEqual(args[0], "Invalid account message: %r")
        self.assert_(isinstance(args[1], Exception))

    @mock.patch('kinto_fxa.scripts.process_account_events.logger')
    def test_invalid_message_is_dropped(self, logger):
        process_account_event(self.config, "{\"Message\": \"{\\\"event\\\": \\\"delete\\\"}\"}")
        calls = logger.exception.call_args_list
        self.assertEqual(len(calls), 1)
        args = calls[0][0]
        self.assertEqual(args[0], "Invalid account message: %r")
        self.assert_(isinstance(args[1], KeyError))

    @mock.patch('kinto_fxa.scripts.process_account_events.logger')
    def test_unknown_message_type_is_dropped(self, logger):
        process_account_event(self.config, json.dumps({
            "Message": json.dumps({
                "event": "rezone",
                "uid": "abcd",
            })
        }))
        logger.warning.assert_called_with(
            "Dropping unknown event type %r",
            "rezone"
        )

    def test_get_default_bucket_id(self):
        self.assertEqual(get_default_bucket_id(self.config, self.uid),
                         self.bucket_id)

    @mock.patch('kinto_fxa.scripts.process_account_events.get_default_bucket_id')
    def test_each_configured_client_is_taken_into_account(self, get_default_bucket_id):
        self.registry.settings = {
            'multiauth.policies': 'ffxxaa',
            'multiauth.policy.ffxxaa.use': 'kinto_fxa.authentication.FxAOAuthAuthenticationPolicy',
            'fxa-oauth.clients.notes.client_id': 'a',
            'fxa-oauth.clients.notes.required_scope': 'a-a',
            'fxa-oauth.clients.lockbox.client_id': 'b',
            'fxa-oauth.clients.lockbox.required_scope': 'b-b',
        }

        process_account_event(self.config, self.real_message)

        user_ids = [c[0][1] for c in get_default_bucket_id.call_args_list]
        assert 'ffxxaa:abcd' in user_ids
        assert 'ffxxaa:abcd-notes' in user_ids
        assert 'ffxxaa:abcd-lockbox' in user_ids

    @mock.patch('kinto_fxa.scripts.process_account_events.get_default_bucket_id')
    def test_valid_message_calls_deletes(self, get_default_bucket_id):
        get_default_bucket_id.return_value = 'some_fxa_bucket'
        process_account_event(self.config, self.real_message)
        self.registry.storage.delete_all.assert_any_call(
            parent_id='/buckets/some_fxa_bucket',
            collection_id=None,
            with_deleted=False,
        )
        self.registry.storage.purge_deleted.assert_any_call(
            parent_id='/buckets/some_fxa_bucket'.format(self.bucket_id),
            collection_id=None,
        )
        self.registry.storage.delete_all.assert_any_call(
            parent_id='/buckets/some_fxa_bucket/*'.format(self.bucket_id),
            collection_id=None,
            with_deleted=False,
        )
        self.registry.storage.purge_deleted.assert_any_call(
            parent_id='/buckets/some_fxa_bucket/*'.format(self.bucket_id),
            collection_id=None,
        )
        self.registry.permission.delete_object_permissions.assert_any_call(
            '/buckets/some_fxa_bucket'.format(self.bucket_id)
        )
        self.registry.permission.delete_object_permissions.assert_any_call(
            '/buckets/some_fxa_bucket/*'.format(self.bucket_id)
        )


class TestProcessAccountEvents(unittest.TestCase):
    def setUp(self):
        boto3_patcher = mock.patch('kinto_fxa.scripts.process_account_events.boto3')
        self.boto3 = boto3_patcher.start()
        self.addCleanup(boto3_patcher.stop)

        self.sqs = mock.Mock()
        self.boto3.resource.return_value = self.sqs

        self.queue = mock.Mock()
        self.sqs.get_queue_by_name.return_value = self.queue

        ec2_metadata_patcher = mock.patch('kinto_fxa.scripts.process_account_events.ec2_metadata')
        self.ec2 = ec2_metadata_patcher.start()
        self.addCleanup(ec2_metadata_patcher.stop)

        self.registry = mock.Mock()
        self.config = {"registry": self.registry}
        self.registry.statsd = None

    @mock.patch('kinto_fxa.scripts.process_account_events.itertools')
    def test_gets_sqs_queue(self, itertools):
        itertools.count.return_value = [1]

        self.queue.receive_messages.return_value = []
        process_account_events(self.config, 'my-queue-name', 'my-aws-region', 23)
        self.boto3.resource.assert_called_with('sqs', region_name='my-aws-region')
        self.sqs.get_queue_by_name.assert_called_with(QueueName='my-queue-name')
        self.queue.receive_messages.assert_called_with(WaitTimeSeconds=23)

    @mock.patch('kinto_fxa.scripts.process_account_events.process_account_event')
    @mock.patch('kinto_fxa.scripts.process_account_events.itertools.count')
    def test_processes_each_message(self, count, process_account_event):
        count.return_value = [1]

        message = mock.Mock(body="my-body")
        self.queue.receive_messages.return_value = [message]

        process_account_events(self.config, 'my-queue-name', 'my-aws-region', 23)
        process_account_event.assert_called_with(self.config, 'my-body')
        message.delete.assert_called_with()

    @mock.patch('kinto_fxa.scripts.process_account_events.logger')
    @mock.patch('kinto_fxa.scripts.process_account_events.process_account_event')
    @mock.patch('kinto_fxa.scripts.process_account_events.itertools.count')
    def test_exceptions_when_processing_messages_are_logged(
            self, count, process_account_event, logger):
        count.return_value = [1]

        message = mock.Mock(body="my-body")
        self.queue.receive_messages.return_value = [message]
        process_account_event.side_effect = ValueError

        with self.assertRaises(ValueError):
            process_account_events(self.config, 'my-queue-name', 'my-aws-region', 23)

        logger.exception.assert_called_with("Error while processing account events")

    @mock.patch('kinto_fxa.scripts.process_account_events.itertools')
    def test_gets_ec2_metadata_if_no_region_given(self, itertools):
        itertools.count.return_value = [1]
        self.ec2.region = 'region-from-metadata'
        self.queue.receive_messages.return_value = []
        process_account_events(self.config, 'my-queue-name', None, 23)
        self.boto3.resource.assert_called_with('sqs', region_name='region-from-metadata')

    @mock.patch('kinto_fxa.scripts.process_account_events.itertools.count')
    def test_uses_statsd_to_send_metrics(self, count):
        count.return_value = [1]

        self.registry.statsd = statsd = mock.Mock()
        process_one = mock.Mock()
        timer_wrapper = mock.Mock()
        timer_wrapper.return_value = process_one
        statsd.timer.return_value = timer_wrapper

        message = mock.Mock(body="my-body")
        self.queue.receive_messages.return_value = [message]

        process_account_events(self.config, 'my-queue-name', 'my-aws-region', 23)
        statsd.timer.assert_called_with('process_account_event')
        timer_wrapper.assert_called_with(process_account_event)
        process_one.assert_called_with(self.config, 'my-body')
        message.delete.assert_called_with()



import gc
import time
import unittest

import mock
import requests
from kinto.core.cache import memory as memory_backend
from kinto.core.testing import DummyRequest
from fxa import errors as fxa_errors
from pyramid import httpexceptions

from kinto_fxa import authentication, DEFAULT_SETTINGS
from kinto_fxa.utils import parse_clients


class TokenVerificationCacheTest(unittest.TestCase):
    def setUp(self):
        cache = memory_backend.Cache(cache_prefix="tests",
                                     cache_max_size_bytes=float("inf"))
        self.cache = authentication.TokenVerificationCache(cache, 0.01)

    def test_set_adds_the_record(self):
        stored = 'toto'
        self.cache.set('foobar', stored)
        retrieved = self.cache.get('foobar')
        self.assertEquals(retrieved, stored)

    def test_delete_removes_the_record(self):
        self.cache.set('foobar', 'toto')
        self.cache.delete('foobar')
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_set_expires_the_value(self):
        self.cache.set('foobar', 'toto')
        time.sleep(0.02)
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_get_ignores_any_error(self):
        with mock.patch.object(self.cache.cache, 'get', side_effect=ValueError):
            retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_delete_ignores_any_error(self):
        self.cache.set('foobar', 'toto')
        with mock.patch.object(self.cache.cache, 'delete', side_effect=ValueError):
            self.cache.delete('foobar')
        retrieved = self.cache.get('foobar')
        self.assertIsNotNone(retrieved)

    def test_set_ignores_any_error(self):
        with mock.patch.object(self.cache.cache, 'set', side_effect=ValueError):
            self.cache.set('foobar', 'toto')
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)


class FxAOAuthAuthenticationPolicyTest(unittest.TestCase):
    def setUp(self):
        self.policy = authentication.FxAOAuthAuthenticationPolicy()
        self.backend = memory_backend.Cache(cache_prefix="tests",
                                            cache_max_size_bytes=float("inf"))

        self.request = self._build_request()
        self.profile_data = {
            "user": "33",
            "scope": ["profile", "mandatory", "optional"],
            "client_id": ""
        }

    def tearDown(self):
        self.backend.flush()

    def _build_request(self):
        request = DummyRequest()
        request.bound_data = {}
        request.registry.cache = self.backend
        settings = DEFAULT_SETTINGS.copy()
        settings['fxa-oauth.cache_ttl_seconds'] = '0.01'
        settings['fxa-oauth.required_scope'] = 'mandatory profile'
        request.registry.settings = settings
        resources, scope_routing = parse_clients(settings)
        request.registry._fxa_oauth_config = resources
        request.registry._fxa_oauth_scope_routing = scope_routing
        request.headers['Authorization'] = 'Bearer foo'
        return request

    def test_returns_none_if_authorization_header_is_missing(self):
        self.request.headers.pop('Authorization')
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIsNone(user_id)

    def test_returns_none_if_token_is_malformed(self):
        self.request.headers['Authorization'] = 'Bearerfoo'
        user_id = self.policy.unauthenticated_userid(self.request)
        self.assertIsNone(user_id)

    def test_returns_none_if_token_is_unknown(self):
        self.request.headers['Authorization'] = 'Carrier foo'
        user_id = self.policy.authenticated_userid(self.request)
        self.assertIsNone(user_id)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid(self, api_mocked):
        api_mocked.return_value = self.profile_data
        user_id = self.policy.authenticated_userid(self.request)
        self.assertEqual("33", user_id)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid_in_principals(self, api_mocked):
        api_mocked.return_value = self.profile_data
        principals = self.policy.effective_principals(self.request)
        self.assertIn("33", principals)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_oauth_verification_is_cached(self, api_mocked):
        # Timing of this test is important. Force a collection to
        # ensure it doesn't happen in between requests.
        gc.collect()
        api_mocked.return_value = self.profile_data
        # First request from client.
        request = self._build_request()
        self.policy.authenticated_userid(request)
        # Second request from same client.
        request = self._build_request()
        self.policy.authenticated_userid(request)
        # Cache backend was used.
        self.assertEqual(1, api_mocked.call_count)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_oauth_verification_is_done_once_per_request(self, api_mocked):
        api_mocked.return_value = self.profile_data
        # First request from client.
        self.policy.authenticated_userid(self.request)
        # Within the same request cycle, token won't be verified.
        self.request.headers['Authorization'] = 'Bearer another'
        self.policy.authenticated_userid(self.request)
        # Request bound data is used.
        self.assertEqual(1, api_mocked.call_count)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_oauth_verification_uses_cache_by_token(self, api_mocked):
        api_mocked.return_value = self.profile_data
        # First request from client.
        request = self._build_request()
        self.policy.authenticated_userid(request)
        # Second request from another client.
        request = self._build_request()
        request.headers['Authorization'] = 'Bearer another'
        self.policy.authenticated_userid(request)
        # Cache backend key was different.
        self.assertEqual(2, api_mocked.call_count)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_oauth_verification_cache_has_ttl(self, api_mocked):
        api_mocked.return_value = self.profile_data
        # First request from client.
        request = self._build_request()
        self.policy.authenticated_userid(request)
        # Second request from same client after TTL.
        time.sleep(0.02)
        request = self._build_request()
        self.policy.authenticated_userid(request)
        # Cache backend key was expired.
        self.assertEqual(2, api_mocked.call_count)

    def test_raise_error_if_oauth2_server_misbehaves(self):
        with mock.patch('kinto_fxa.authentication.'
                        'OAuthClient.verify_token') as mocked:
            mocked.side_effect = fxa_errors.OutOfProtocolError
            self.assertRaises(httpexceptions.HTTPServiceUnavailable,
                              self.policy.authenticated_userid,
                              self.request)

    def test_returns_none_if_oauth2_error(self):
        with mock.patch('kinto_fxa.authentication.'
                        'OAuthClient.verify_token') as mocked:
            mocked.side_effect = fxa_errors.ClientError
            self.assertIsNone(self.policy.authenticated_userid(self.request))

    def test_returns_none_if_oauth2_scope_mismatch(self):
        with mock.patch('kinto_fxa.authentication.'
                        'OAuthClient.verify_token') as mocked:
            mocked.side_effect = fxa_errors.TrustError
            self.assertIsNone(self.policy.authenticated_userid(self.request))

    def test_forget_uses_realm(self):
        policy = authentication.FxAOAuthAuthenticationPolicy(realm='Who')
        headers = policy.forget(self.request)
        self.assertEqual(headers[0],
                         ('WWW-Authenticate', 'Bearer realm="Who"'))


class FxAPingTest(unittest.TestCase):
    def setUp(self):
        self.request = DummyRequest()
        self.request.registry.settings = DEFAULT_SETTINGS
        self.request.registry.settings['fxa-oauth.oauth_uri'] = 'http://fxa'

    def test_returns_none_if_oauth_deactivated(self):
        self.request.registry.settings['fxa-oauth.oauth_uri'] = None
        self.assertIsNone(authentication.fxa_ping(self.request))

    @mock.patch('requests.get')
    def test_returns_true_if_ok(self, get_mocked):
        httpOK = requests.models.Response()
        httpOK.status_code = 200
        get_mocked.return_value = httpOK
        self.assertTrue(authentication.fxa_ping(self.request))

    @mock.patch('requests.get')
    def test_returns_false_if_ko(self, get_mocked):
        get_mocked.side_effect = requests.exceptions.HTTPError()
        self.assertFalse(authentication.fxa_ping(self.request))


class FxAOAuthAuthenticationMultipleClientsPolicyTest(unittest.TestCase):
    def setUp(self):
        self.policy = authentication.FxAOAuthAuthenticationPolicy()
        self.backend = memory_backend.Cache(cache_prefix="tests",
                                            cache_max_size_bytes=float("inf"))

        self.request = self._build_request()

        self.profile_data = {
            "user": "33",
            "scope": ["profile", "https://identity.mozilla.org/apps/notes"],
            "client_id": "c73e46074a948932"
        }

    def tearDown(self):
        self.backend.flush()

    def _build_request(self):
        request = DummyRequest()
        request.bound_data = {}
        request.registry.cache = self.backend
        settings = DEFAULT_SETTINGS.copy()
        settings['fxa-oauth.oauth_uri'] = 'https://oauth.accounts.firefox.com/v1'
        settings['fxa-oauth.cache_ttl_seconds'] = '0.01'
        settings['fxa-oauth.clients.notes.client_id'] = 'c73e46074a948932'
        settings['fxa-oauth.clients.notes.required_scope'] = (
            'profile https://identity.mozilla.org/apps/notes')
        settings['fxa-oauth.clients.lockbox.client_id'] = '299062f8b3838932'
        settings['fxa-oauth.clients.lockbox.required_scope'] = (
            'profile https://identity.mozilla.org/apps/lockbox')

        request.registry.settings = settings
        resources, scope_routing = parse_clients(settings)
        request.registry._fxa_oauth_config = resources
        request.registry._fxa_oauth_scope_routing = scope_routing
        request.headers['Authorization'] = 'Bearer foo'
        return request

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid_for_notes(self, api_mocked):
        api_mocked.return_value = self.profile_data
        user_id = self.policy.authenticated_userid(self.request)
        self.assertEqual("33-notes", user_id)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid_in_principals_for_notes(self, api_mocked):
        api_mocked.return_value = self.profile_data
        principals = self.policy.effective_principals(self.request)
        self.assertIn("fxa:33", principals)
        self.assertNotIn("33", principals)
        self.assertIn("33-notes", principals)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_fails_to_match_a_client_if_only_one_of_the_required_scopes(self, api_mocked):
        api_mocked.return_value = {
            "user": "33",
            "scope": ["profile"],
            "client_id": ""
        }
        user_id = self.policy.authenticated_userid(self.request)
        self.assertEqual(user_id, None)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid_for_lockbox(self, api_mocked):
        api_mocked.return_value = {
            "user": "33",
            "scope": ["profile", "https://identity.mozilla.org/apps/lockbox"],
            "client_id": ""
        }
        user_id = self.policy.authenticated_userid(self.request)
        self.assertEqual("33-lockbox", user_id)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_returns_fxa_userid_in_principals_for_lockbox(self, api_mocked):
        api_mocked.return_value = {
            "user": "33",
            "scope": ["profile", "https://identity.mozilla.org/apps/lockbox"],
            "client_id": ""
        }
        principals = self.policy.effective_principals(self.request)
        self.assertIn("fxa:33", principals)
        self.assertNotIn("33", principals)
        self.assertIn("33-lockbox", principals)

    @mock.patch('fxa.oauth.APIClient.post')
    def test_fails_to_connect_if_bearer_matches_multiple_config(self, api_mocked):
        api_mocked.return_value = {
            "user": "33",
            "scope": ["profile", "https://identity.mozilla.org/apps/notes",
                      "https://identity.mozilla.org/apps/lockbox"],
            "client_id": ""
        }
        principals = self.policy.effective_principals(self.request)
        self.assertNotIn("fxa:33", principals)
        self.assertNotIn("33-lockbox", principals)
        self.assertNotIn("33-notes", principals)



"""
Main entry point for kinto_fxa scripts.
"""

import argparse
import logging.config
import os

from pyramid.paster import bootstrap

from .process_account_events import process_account_events

DEFAULT_CONFIG_FILE = os.getenv('KINTO_INI', 'config/kinto.ini')
logger = logging.getLogger(__name__)


def main(args=None):
    parser = argparse.ArgumentParser(description="Listen to the queue for account messages.")
    parser.add_argument('--ini', dest='ini_file', required=False, default=DEFAULT_CONFIG_FILE,
                        help="path to kinto INI file")
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='kinto-fxa command to run',
                                       dest='subcommand')
    subparsers.required = True
    subparser = subparsers.add_parser('process-account-events')

    subparser.add_argument('queue_name',
                           help="SQS queue on which to listen for events")
    subparser.add_argument('--aws-region',
                           help="aws region in which the queue can be found")
    subparser.add_argument("--queue-wait-time", type=int, default=20,
                           help="Number of seconds to wait for jobs on the queue")

    opts = parser.parse_args(args)

    logging.config.fileConfig(opts.ini_file, disable_existing_loggers=False)
    logger.debug("Using config file %r", opts.ini_file)
    config = bootstrap(opts.ini_file)

    process_account_events(
        config, opts.queue_name,
        opts.aws_region, opts.queue_wait_time)
    return 0


if __name__ == '__main__':  # pragma: nocover
    main()



import logging
from urllib.parse import urljoin

import requests
from fxa.oauth import Client as OAuthClient
from fxa import errors as fxa_errors
from pyramid import authentication as base_auth
from pyramid import httpexceptions
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.settings import aslist
from zope.interface import implementer

from kinto_fxa.utils import fxa_conf

logger = logging.getLogger(__name__)

REIFY_KEY = 'fxa_verified_token'


class TokenVerificationCache(object):
    """Verification cache class as expected by PyFxa library.

    This basically wraps the cache backend instance to specify a constant ttl.
    """
    def __init__(self, cache, ttl):
        self.cache = cache
        self.ttl = ttl

    def get(self, key):
        try:
            return self.cache.get(key)
        except Exception:
            logger.exception("Error while fetching from cache")
        return None

    def set(self, key, value):
        try:
            self.cache.set(key, value, self.ttl)
        except Exception:
            logger.exception("Error while storing in cache")

    def delete(self, key):
        try:
            self.cache.delete(key)
        except Exception:
            logger.exception("Error while deleting from cache")


@implementer(IAuthenticationPolicy)
class FxAOAuthAuthenticationPolicy(base_auth.CallbackAuthenticationPolicy):
    def __init__(self, realm='Realm'):
        self.realm = realm
        self._cache = None
        self._auth_client = None

    def unauthenticated_userid(self, request):
        """Return the FxA userid or ``None`` if token could not be verified.
        """
        authorization = request.headers.get('Authorization', '')
        try:
            authmeth, token = authorization.split(' ', 1)
        except ValueError:
            return None
        if authmeth.lower() != 'bearer':
            return None

        user_id, client_name = self._verify_token(token, request)

        # Don't add suffix if authentication failed, or no specific client name is configured
        if client_name is None or client_name == 'default':
            return user_id

        return '{}-{}'.format(user_id, client_name)

    def forget(self, request):
        """A no-op. Credentials are sent on every request.
        Return WWW-Authenticate Realm header for Bearer token.
        """
        return [('WWW-Authenticate', 'Bearer realm="%s"' % self.realm)]

    def _verify_token(self, token, request):
        """Verify the token extracted from the Authorization header.

        This method stores the result in two locations to avoid hitting the
        auth remote server as much as possible:

        - on the request object, in case the Pyramid authentication methods
          like `effective_principals()` or `authenticated_userid()` are called
          several times during the request cycle;

        - in the cache backend, to reuse validated token from one request to
          another (during ``cache_ttl_seconds`` seconds.)
        """
        # First check if this request was already verified.
        # `request.bound_data` is an attribute provided by Kinto to store
        # some data that is shared among sub-requests (e.g. default bucket
        # or batch requests)
        if REIFY_KEY not in request.bound_data:
            user_id = None
            client_name = None
            auth_client = self._get_auth_client(request)

            for scope, client in request.registry._fxa_oauth_scope_routing.items():
                try:
                    profile = auth_client.verify_token(token=token, scope=aslist(scope))
                    user_id = profile['user']
                    scope = profile['scope']
                    client_name = client

                    # Make sure the bearer token scopes don't match multiple configs.
                    routing_scopes = request.registry._fxa_oauth_scope_routing
                    intersecting_scopes = [x for x in routing_scopes.keys()
                                           if x and set(x.split()).issubset(set(scope))]
                    if len(intersecting_scopes) > 1:
                        logger.warn("Invalid FxA token: {} matches multiple config" % scope)
                        return None, None

                    break
                except fxa_errors.OutOfProtocolError:
                    logger.exception("Protocol error")
                    raise httpexceptions.HTTPServiceUnavailable()
                except (fxa_errors.InProtocolError, fxa_errors.TrustError) as e:
                    logger.debug("Invalid FxA token: %s" % e)

            # Save for next call.
            request.bound_data[REIFY_KEY] = (user_id, client_name)

        return request.bound_data[REIFY_KEY]

    def _get_cache(self, request):
        """Instantiate cache when first request comes in.
        This way, the policy instantiation is decoupled from registry object.
        """
        if self._cache is None:
            if hasattr(request.registry, 'cache'):
                cache_ttl = float(fxa_conf(request, 'cache_ttl_seconds'))
                oauth_cache = TokenVerificationCache(request.registry.cache,
                                                     ttl=cache_ttl)
                self._cache = oauth_cache

        return self._cache

    def _get_auth_client(self, request):
        """Instantiate OAuthClient on first request but cache it.

        Hopefully this lets us retain the requests Session in the
        PyFxA class and keep the HTTP connection alive for longer.
        """
        if self._auth_client is None:
            # Use PyFxa defaults if not specified
            server_url = fxa_conf(request, 'oauth_uri')
            auth_cache = self._get_cache(request)
            self._auth_client = OAuthClient(server_url=server_url, cache=auth_cache)

        return self._auth_client

    def callback(self, userid, request):
        if request.bound_data.get(REIFY_KEY, (None, "default"))[1] != "default":
            # Add the usual FxA ID as a principal
            user_id = request.bound_data[REIFY_KEY][0]
            return ["fxa:{}".format(user_id)]
        return []


def fxa_ping(request):
    """Verify if the OAuth server is ready."""
    server_url = fxa_conf(request, 'oauth_uri')

    oauth = None
    if server_url is not None:
        auth_client = OAuthClient(server_url=server_url)
        server_url = auth_client.server_url
        oauth = False

        try:
            heartbeat_url = urljoin(server_url, '/__heartbeat__')
            timeout = float(fxa_conf(request, 'heartbeat_timeout_seconds'))
            r = requests.get(heartbeat_url, timeout=timeout)
            r.raise_for_status()
            oauth = True
        except requests.exceptions.HTTPError:
            pass

    return oauth
